WINDOW_WIDTH = 1800
WINDOW_HEIGHT = 960 

ONED_WIDTH = 900
ONED_HEIGHT = 900

THREED_WIDTH = 500
THREED_HEIGHT = 500

INFO_WIDTH = 400

SCROLL_WIDTH = 260
THUMBNAIL_WIDTH = 200

BG_COLOR = "#0B1117 "
LIGHT_BG_COLOR = "#131B2A"
VERY_LIGHT_BG_COLOR = "#507BA6"
BORDER_COLOR = "#243242"
LIGHT_BORDER_COLOR = "#374558"
TEXT_COLOR = "#E0E1DD"

